import os
import argparse

# Reading arguments from command line
ap = argparse.ArgumentParser(description = "Collecting all PSIBLAST hits in specified folder and sorting them between clusters. Marking hits that are nearby seeds and storing distance to the seeds.")
ap.add_argument("-c", help = "Folder name where hits stored", required = True)
ap.add_argument("-o", help = "Folder name where sorted result hits stored", required = True)
ap.add_argument("-p", help = "PTY file containing coordinates of ORFs of genomic database", required = True)
ap.add_argument("-i", help = "List of protein IDs in vicinity of baits", required = True)
ap.add_argument("-s", help = "Baits coordinates", required = True)
ap.add_argument("-v", help = "Vicinity around the baits, needed to calculate distances to the baits", required = True)
ap.add_argument("-z", default=0.4, help = "Overlap threshold, hits are subject to sort between two profiles if they are overlapping for more than threshold value", required = False)
ap.add_argument("-x", default=0.25, help = "Coverage threshold, hits are stored if they cover original profile for more than threshold value", required = False)

opts = ap.parse_args()
ClustersHitsFolder = opts.c
SortedHitsFolder = opts.o
PTYFileName = opts.p
VicinityProteinIDsFileName = opts.i
SeedsFileName = opts.s
VicinityFileName = opts.v
MIN_COVERAGE_FILTER = float(opts.x)
MIN_OVERLAP_FILTER = float(opts.z)


# MIN_COVERAGE_FILTER = 0.25
# MIN_OVERLAP_FILTER = 0.4
# ClustersHitsFolder = '/mnt/d/lsd/Icity/CLUSTERS_DND_A/'
# SortedHitsFolder =  '/mnt/d/lsd/Icity/CLUSTERS_DND_A/Sorted/'
# PTYFileName = '/mnt/d/lsd/Icity/Database/archaea_cds.pty'
# VicinityProteinIDsFileName = '/mnt/d/lsd/Icity/VicinityIDs_DND_A.lst'
# SeedsFileName = '/mnt/d/lsd/Icity/Seeds_DND_A.tsv'
# VicinityFileName = '/mnt/d/lsd/Icity/Vicinity_DND_A.tsv'

HIT_Score = 1
HIT_Cluster = 5
HIT_ALIGNMENT_START = 2
HIT_ALIGNMENT_STOP = 3
HIT_ALIGNMENT_Sequence = 4
HIT_Protein_ID = 0

INFO_CONTIG = 0
INFO_BAITED = 1
INFO_ORF_START = 2
INFO_ORF_STOP = 3
INFO_ORF_DISTANCE_TO_BAIT = 4


def AddHitToDict(TargetHitsDict, Hit):
    BetterHitExists = False

    for DictHit in TargetHitsDict[Hit[HIT_Protein_ID]]:
        InterCoverage = min(Hit[HIT_ALIGNMENT_STOP],
                            DictHit[HIT_ALIGNMENT_STOP] - max(Hit[HIT_ALIGNMENT_START], DictHit[HIT_ALIGNMENT_START]))
        MinLength = min(Hit[HIT_ALIGNMENT_STOP] - Hit[HIT_ALIGNMENT_START],
                        DictHit[HIT_ALIGNMENT_STOP] - DictHit[HIT_ALIGNMENT_START])

        if InterCoverage / MinLength > MIN_COVERAGE_FILTER:
            if Hit[HIT_Score] > DictHit[HIT_Score]:
                TargetHitsDict[Hit[HIT_Protein_ID]].remove(DictHit)
            else:
                BetterHitExists = True
                break

    if not BetterHitExists:
        TargetHitsDict[Hit[HIT_Protein_ID]].append(Hit)

    return TargetHitsDict



def AddTargetHitsValues(TargetHitsFileName, TargetHitsDict, ProteinInfoDict):  # TargetHitsDict):
    BLAST_FORMAT_ALIGNMENT_TARGET_PROTEIN_ID = 1
    BLAST_FORMAT_ALIGNMENT_START = 8
    BLAST_FORMAT_ALIGNMENT_STOP = 9
    BLAST_FORMAT_ALIGNMENT_SCORE = 10
    BLAST_FORMAT_ALIGNMENT_SEQUENCE = 7

    TargetValues = []
    BestScoreLine = []
    MaxScore = 0
    ClusterId = os.path.splitext(os.path.basename(TargetHitsFileName))[0]

    with open(TargetHitsFileName, "r") as TargetHitsFile:
        for Line in TargetHitsFile:
            if Line[0] == "#":
                continue

            LineValues = Line.split("\t")
            LineValues[-1] = LineValues[-1][:-1]

            # Alignment Data
            Start = int(LineValues[BLAST_FORMAT_ALIGNMENT_START])
            Stop = int(LineValues[BLAST_FORMAT_ALIGNMENT_STOP])
            Score = int(LineValues[BLAST_FORMAT_ALIGNMENT_SCORE])
            if "|" in LineValues[BLAST_FORMAT_ALIGNMENT_TARGET_PROTEIN_ID]:
                ProteinID = LineValues[BLAST_FORMAT_ALIGNMENT_TARGET_PROTEIN_ID].split("|")[
                    1]  # Accession|ID| format expected
            else:
                ProteinID = LineValues[BLAST_FORMAT_ALIGNMENT_TARGET_PROTEIN_ID]

            HitLine = [ProteinID, Score, Start, Stop, LineValues[BLAST_FORMAT_ALIGNMENT_SEQUENCE], ClusterId]
            HitLine.extend(ProteinInfoDict[ProteinID])

            if Score > MaxScore:
                BestScoreLine = HitLine
                MaxScore = Score

            TargetValues.append(HitLine)

    if len(TargetValues) > 0:  # no hits
        # max length of alignment is taken as selflength to filter by coverage
        # this value is used as representative length of alignment
        BestScoreLength = len(BestScoreLine[HIT_ALIGNMENT_Sequence].replace("-", ""))

        for Hit in TargetValues:
            if len(Hit[HIT_ALIGNMENT_Sequence].replace("-",
                                                       "")) / BestScoreLength >= MIN_OVERLAP_FILTER:  # filtering by coverage
                if Hit[HIT_Protein_ID] in TargetHitsDict:
                    TargetHitsDict = AddHitToDict(TargetHitsDict, Hit)
                else:
                    TargetHitsDict[Hit[HIT_Protein_ID]] = [Hit]

    return ClusterId

def GetDistToSeedDict(AnnotationFileName, SeedsDict):
    LOCI_PROTEIN_ID = 0
    LOCI_CONTIG_ID = 4
    LOCI_COORDINATES = 1

    DistToSeed = dict()

    Accessions = dict()
    IslandSeeds = []
    for Line in open(AnnotationFileName, "r"):
        LineValues = Line[:-1].split("\t")

        if LineValues[LOCI_PROTEIN_ID] == "===":
            if len(Accessions) > 0:
                if len(IslandSeeds) == 0:
                    raise "NoSeedFound"
                else:
                    for Accession in Accessions:
                        DistToSeed[Accession] = min([abs(x - Accessions[Accession]) for x in IslandSeeds])

            Accessions = dict()
            IslandSeeds = []
        else:
            Accessions[LineValues[LOCI_PROTEIN_ID]] = len(Accessions) + 1

            StartStop = LineValues[LOCI_COORDINATES].split("..")
            if IsInSeeds(LineValues[LOCI_CONTIG_ID], int(StartStop[0]), int(StartStop[1]), SeedsDict):
                IslandSeeds.append(len(Accessions))

    if len(Accessions) > 0:
        if len(IslandSeeds) == 0:
            raise "NoSeedFound"
        else:
            for Accession in Accessions:
                DistToSeed[Accession] = min([abs(x - Accessions[Accession]) for x in IslandSeeds])

    return DistToSeed

def LoadProteinInfoDict(PTYFileName, VicinityProteinIDs, LociDists):
    PTY_PROTEIN_ID = 5
    PTY_CONTIG_ID = 4
    PTY_COORDINATES = 1

    ProteinInfoDict = dict()

    for Line in open(PTYFileName):
        LineValues = Line[:-1].split("\t")

        StartStop = LineValues[PTY_COORDINATES].split(":")

        if LineValues[PTY_PROTEIN_ID] in VicinityProteinIDs:
            InVicinity = "1"
        else:
            InVicinity = "0"

        if LineValues[PTY_PROTEIN_ID] in LociDists:
            Dist = LociDists[LineValues[PTY_PROTEIN_ID]]
        else:
            Dist = 10000

        # in assumption that Protein IDs are unique
        ProteinInfoDict[LineValues[PTY_PROTEIN_ID]] = [LineValues[PTY_CONTIG_ID], InVicinity,
                                                       StartStop[0], StartStop[1], Dist]

    return ProteinInfoDict

def LoadList(FileName):
    ListValues = []
    for Line in open(FileName):
        ListValues.append(Line[:-1])
    return ListValues

def LoadSeedsDict(SeedsFileName):
    Seeds = dict()
    ContigField = 3
    StartField = 4
    EndField = 5
    Offset = 0

    return AddSeedsDictByFields(SeedsFileName, Seeds, ContigField, StartField, EndField, Offset)

def AddSeedsDictByFields(FileName, SeedDict, ContigField, StartField, EndField, Offset):
    with open(FileName, "r") as File:
        for Line in File:
            LineValues = Line.split("\t")
            LineValues[-1] = LineValues[-1][:-1]
            ID = LineValues[0]

            Start, End = minmax(int(LineValues[StartField]), int(LineValues[EndField]))
            if LineValues[ContigField] in SeedDict:
                SeedDict[LineValues[ContigField]].append([max(Start - Offset, 0), End + Offset, ID, LineValues])
            else:
                SeedDict[LineValues[ContigField]] = [[max(Start - Offset, 0), End + Offset, ID, LineValues]]

    for Key in SeedDict:
        SeedDict[Key] = sorted(SeedDict[Key], key = lambda e: e[0])

    return SeedDict

def minmax(X, Y):
    return min(X, Y), max(X, Y)

def IsInSeeds(ContiAccessionD, Start, Stop, SeedsDict):
    Seed = GetSeed(ContiAccessionD, Start, Stop, SeedsDict)
    if len(Seed) == 0:
        return False
    return True

def GetSeed(ContiAccessionD, Start, Stop, SeedsDict):
    if not ContiAccessionD in SeedsDict:
        return []

    for ORF in SeedsDict[ContiAccessionD]:
        if (Start <= ORF[1]) and (Stop >= ORF[0]):
            return ORF

    return []

VicinityProteinIDs = set(LoadList(VicinityProteinIDsFileName))
Seeds = LoadSeedsDict(SeedsFileName)
LociDists = GetDistToSeedDict(VicinityFileName, Seeds)
ProteinInfoDict = LoadProteinInfoDict(PTYFileName, VicinityProteinIDs, LociDists)

if not os.path.exists(SortedHitsFolder):
    os.makedirs(SortedHitsFolder)

ProteinHitsDict = dict()
Clusters = []
for FileName in os.listdir(ClustersHitsFolder):
    if not FileName.endswith(".hits"):
        continue

    ClusterId = AddTargetHitsValues(ClustersHitsFolder + FileName, ProteinHitsDict, ProteinInfoDict)
    Clusters.append(ClusterId)

print("Loading BLAST hits complete.")

ClusterHitsDict = dict()
for ProteinID in ProteinHitsDict:
    for Hit in ProteinHitsDict[ProteinID]:
        if Hit[HIT_Cluster] in ClusterHitsDict:
            ClusterHitsDict[Hit[HIT_Cluster]].append(Hit)
        else:
            ClusterHitsDict[Hit[HIT_Cluster]] = [Hit]

print("Moving sorting hits by clusters complete.")

for ClusterId in Clusters:
    if ClusterId in ClusterHitsDict:
        with open(SortedHitsFolder + str(ClusterId) + ".hits_sorted", "w") as SortedFile:
            for Hit in ClusterHitsDict[ClusterId]:
                ResLine = "\t".join([str(x) for x in Hit])

                SortedFile.write(ResLine + "\n")
    else: # empty file
        with open(SortedHitsFolder + str(ClusterId) + ".hits_sorted", "w") as SortedFile:
            SortedFile.write("")



#Default configurations

"""
Configs for @Icity Runner.
[Archaea_RM] processing in default, which can be changed by requirements.
"""
# Input files and Hyper-parameters
import os
DefenseSystem_Name = "RM_A"
DefenseSystem_FilePath = "/root/ICityRunnerPackage/"  # Your Working Path

ICITY_CONFIG_INPUT = {
    "PTYFile": "Database/archaea_cds.pty",
    "PathToDatabase": "Database/ArchaeaProt",
    # "PTYFile": "DatabaseDEMO/CDS.pty", # DEMO
    # "PathToDatabase": "DatabaseDEMO/ProteinsDB", # DEMO
    # "SeedsFile": "Seeds.tsv",
    "CdsPath": "Database/archaea_cds.pty",
    "SeedPath": "Archaea_RM.xlsx",
    "NeighborhoodVicinitySize": 10000,
    "PermissiveClusteringThreshold": 0.3,
    "SortingOverlapThreshold": 0.4,
    "SortingCoverageThresold": 0.25,
    "ThreadNum": "48"   # Change global thread number here
}

# Output files
ICITY_CONFIG_OUTPUT = {
    "ICITYFileName": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "Relevance_"+DefenseSystem_Name+".tsv"),
    "VicinityClustersFileName": "VicinityPermissiveClustsLinear" + DefenseSystem_Name + ".tsv",
    "RelevanceCategoryName": "RelevanceCategory_" + DefenseSystem_Name + ".csv"
}

# Temporary files
ICITY_CONFIG_TEMPORARYFILES = {
    "VicinityFileName": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "Vicinity_"+DefenseSystem_Name+".tsv"),
    "VicinityIDsFileName": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "VicinityIDs_"+DefenseSystem_Name+".lst"),
    "VicinityFASTAFileName": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "Vicinity_"+DefenseSystem_Name+".faa"),
    "VicinityClustersFileName": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "VicinityPermissiveClustsLinear_"+DefenseSystem_Name+".tsv"),
    "ProfilesFolder": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "CLUSTERS_"+DefenseSystem_Name + "/"),
    "SortedBLASTHitsFolder": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "CLUSTERS_"+DefenseSystem_Name, "Sorted/"),
    "SeedsExtractedFileName": os.path.join("./" + DefenseSystem_Name+"_OUTPUT", "Seeds_" + DefenseSystem_Name + ".tsv"),
    # "SeedsExtractedFileName": "Seeds.tsv"   # DEMO

}



import argparse

ap = argparse.ArgumentParser(description = "Removes everything except first ID from FASTA ID line")
ap.add_argument("-f", help = "FASTA file name", required = True)

opts = ap.parse_args()
FASTAFileName = opts.f

for Line in open(FASTAFileName):
    if ">" in Line:
        FullID = Line[1:-1].split(" ")[0]
        ShortID = FullID
        if "|" in ShortID:
            ShortID = ShortID.split("|")[1]

        print(">" + ShortID)
    else:
        print(Line[:-1])


import os
import subprocess
import argparse

def GetNumberOfSequences(FASTAFileName):
    Count = 0
    for Line in open(FASTAFileName):
        if ">" in Line:
            Count += 1

    return Count

ap = argparse.ArgumentParser(description = "Running PSIBLAST for each protein profile in specified folder and saving results in same folder")
ap.add_argument("-c", help = "Folder name where profiles stored", required = True)
ap.add_argument("-d", help = "Path to protein database", required = True)
ap.add_argument("-t", help = "Thread Number", required = True)


opts = ap.parse_args()
ClustersFolderName = opts.c
Database = opts.d
ThreadNum = opts.t

Evalue = 0.000001

for FileName in os.listdir(ClustersFolderName):
    if not FileName.endswith("ali"):
        continue

    BLASTHitsFileName = ClustersFolderName + "/" + os.path.splitext(FileName)[0] + ".hits"

    if GetNumberOfSequences(ClustersFolderName + "/" + FileName) == 1:
        Query = "-query"
    else:
        Query = "-in_msa"
    Query += " " + ClustersFolderName + "/" + FileName

    # to parallelize
    print("Processing " + FileName)
    subprocess.call(
        "psiblast -db " + Database + " " +
        "-outfmt \"7 qseqid sseqid slen sstart send evalue qseq sseq qstart qend score\"" +
        " -seg no -evalue " + str(Evalue) +
        " -dbsize 20000000 -max_target_seqs 10000 -comp_based_stats no " +
        "-num_threads " + ThreadNum +
        Query + " > " + BLASTHitsFileName, shell = True)



import argparse
import subprocess
import os
import time
from multiprocessing import Pool

ap = argparse.ArgumentParser(description = "Creates protein profiles in specified folder for clusters in given file")
ap.add_argument("-f", help = "Clusters file name", required = True)
ap.add_argument("-c", help = "Folder name where profiles will be saved", required = True)
ap.add_argument("-d", help = "Path to protein database", required = True)

opts = ap.parse_args()
ClustersFileName = opts.f
ClustersFolderName = opts.c
Database = opts.d

ClusterNo = 0
TmpIDsFileName = "Tmp_IDs.lst"
TmpFASTAFileName = "Tmp_FASTA.faa"

if not os.path.exists(ClustersFolderName):
    subprocess.call("mkdir " + ClustersFolderName, shell = True)

def profile_worker(ClusterNo, Line, Length):
    global ClustersFileName
    print("Processing No: ",'[', ClusterNo, '/', Length,']', "Sequence: ", Line)
    ClusterIDs = Line[:-1].split("\t")[1].split(" ")
    ClusterProfileFileName = "CLUSTER_" + str(ClusterNo) + ".ali"

    with open(TmpIDsFileName, "w") as IDsFile:
        IDsFile.write("\n".join(ClusterIDs))

    subprocess.call("blastdbcmd -db " + Database + " -entry_batch " + TmpIDsFileName + " -long_seqids > " + TmpFASTAFileName,
                    shell=True)

    subprocess.call("muscle -align " + TmpFASTAFileName + " -output " + ClustersFolderName + "/" + ClusterProfileFileName
                    , shell=True)


def multi_process():
    cpu_cnt = os.cpu_count()
    print("CPU Core Num: ", cpu_cnt)
    pool = Pool(cpu_cnt)  # Creating Process Pool
    with open(ClustersFileName, 'r') as File:
        Lines = File.readlines(); Length = len(Lines)
        for ClusterNo, Line in enumerate(Lines):
            print("No.", ClusterNo, "Processing sequence: ", Line)
            pool.apply_async(profile_worker, (ClusterNo + 1, Line, Length))  # Non-Blocking

    pool.close()  # Close Proces Pool Entry
    pool.join()  # Waiting
    print("Work Done !")


if __name__ == '__main__':
    t1 = time.time()
    cnt = 0
    multi_process()

    t2 = time.time()

    subprocess.call("rm " + TmpIDsFileName, shell=True)
    subprocess.call("rm " + TmpFASTAFileName, shell=True)

    print("Used time: ", t2 - t1, "s")







import argparse
import re
def LoadSeedDirt(SeedFileName, NoID =False):
    Seeds = dict()

    ContigID = 3
    Start = 4
    End = 5
    Offset = 0

    return AddSeedDictFields(SeedFileName, Seeds, ContigID, Start, End, Offset, NoID)

def AddSeedDictFields(SeedFileName, Seeds, ContigID, Start, End, Offset, NoID=False):
    with open(SeedFileName, 'r') as File:
        for Line in File:
            LineValues = Line.split('\t')
            LineValues[-1] = LineValues[-1][:-1]

            if not NoID:
                ID = LineValues[2]
            else:
                ID = 'Seed'

            ORF_start, ORF_end = minmax(int(LineValues[Start]), int(LineValues[End]))
            if LineValues[ContigID] in Seeds:
                Seeds[LineValues[ContigID]].append([max(ORF_start - Offset, 0), ORF_end + Offset, ID, LineValues])
            else:
                Seeds[LineValues[ContigID]] = [[max(ORF_start - Offset, 0), ORF_end + Offset, ID, LineValues]]
            
        for Key in Seeds:
            Seeds[Key] = sorted(Seeds[Key], key= lambda e:e[0])
    return Seeds

def minmax(X, Y):
    return min(X, Y), max(X, Y)

# CDS PTY tsv file columns
PTY_LocusTag = 0
PTY_Coordinates = 1
PTY_Strand = 2
PTY_SpecieName = 3
PTY_ContigId = 4
PTY_AccessionNo = 5

def GetSeed(ContigID, Start, End, Seeds):
    if not ContigID in Seeds:
        return []
    
    for ORF in Seeds[ContigID]:
        if(Start <= ORF[1]) and (End >= ORF[0]):
            return ORF

    return []

def IsInSeeds(ContigID, Start, End, Seeds):
    InSeed = GetSeed(ContigID, Start, End, Seeds)
    if len(InSeed) == 0:
        return False
    return True

def WriteSeedIslands(SeedIslandPTYLines, ResultFile):

    if len(SeedIslandPTYLines) > 0:
        ResultFile.write('===\n')
        for LineValues in SeedIslandPTYLines:
            ResultFile.write(
            LineValues[0] + '\t' + 
            LineValues[1].replace(':','..') + '\t' +
            LineValues[2] + '\t' + 
            LineValues[3] + '\t' +
            LineValues[4] + '\n')


def WriteIslands(ContigLines, ResultFile, ContigID, Seeds, SeedIslandOffset):
    if ContigID not in Seeds:
        return
    if len(Seeds[ContigID]) == 0:
        return 
    
    
    ContigLines = sorted(ContigLines, key= lambda e:e[0])

    SeedIslandPTYLines = []
    for Line in ContigLines:
        LineValues = Line[1]

        Coordinates = LineValues[PTY_Coordinates].split(":")
        
        
        try:
            Start = int(Coordinates[0])
            End = int(Coordinates[1])
        except ValueError:
            continue

        if IsInSeeds(ContigID, Start - SeedIslandOffset, End + SeedIslandOffset, Seeds):
            SeedIslandPTYLines.append(
                [LineValues[PTY_AccessionNo], LineValues[PTY_Coordinates], LineValues[PTY_Strand],
                LineValues[PTY_SpecieName], LineValues[PTY_ContigId]])
        else:
            if len(SeedIslandPTYLines) > 0:
                WriteSeedIslands(SeedIslandPTYLines, ResultFile)

            SeedIslandPTYLines = []
    
    if len(SeedIslandPTYLines) > 0:
        WriteSeedIslands(SeedIslandPTYLines, ResultFile)

if __name__ == "__main__":
    # Reading arguments from command line
    ap = argparse.ArgumentParser(description="Extracting seeds from CDS file")
    ap.add_argument("-c", help="CDSFileName, cds pty file, containing seed information", required=True)
    ap.add_argument("-s", help="SeedsFileName, seeds tsv file", required=True)
    ap.add_argument("-o", help="ResultFileName, output tsv file", required=True)
    ap.add_argument("-d", help="Offset around seed (base pairs)", required=True)
    opts = ap.parse_args()

    SeedFileName = opts.s  # Archaea_Seeds_RM_A.tsv
    PTYDataFileName = opts.c  # archaea_cds.pty
    ResultFileName = opts.o  # Vicinity.tsv

    SeedIslandOffset = int(opts.d)
    Seeds = LoadSeedDirt(SeedFileName)

    ContigID = ''
    ContigLines = []
    with open(ResultFileName, 'w') as ResultFile:
        with open(PTYDataFileName,'r') as PTYData:
            for Line in PTYData:
                Line = Line[:-1]
                LineValues = Line.split('\t')

                if ContigID != LineValues[PTY_ContigId]:
                    WriteIslands(ContigLines, ResultFile, ContigID, Seeds, SeedIslandOffset)
                    ContigLines =[]

                ContigID = LineValues[PTY_ContigId]
                Coordinates = LineValues[PTY_Coordinates].split(':')
                Coordinates[0] = re.findall(r'\d+', Coordinates[0])[0]
                Start = int(Coordinates[0])


                ContigLines.append([Start, LineValues])

            WriteIslands(ContigLines, ResultFile, ContigID, Seeds, SeedIslandOffset)



import config
import subprocess
import argparse
import shutil
import os
import sys


def subprocess_call(Stage, CallText):

    print(Stage)
    print(CallText)
    try:
        # print("Calling: ", CallText)
        devnull = open(os.devnull, 'wb')
        subprocess.check_call(CallText, shell=True, stdout=devnull, stderr=devnull)
        '''
        # Fetch print std
        process = subprocess.Popen(CallText, shell=True, stdout=subprocess.PIPE,stderr=subprocess.PIPE )
        while True:
            if process.poll() is None:
                line = process.stdout.readline().strip()
                print(line)
       '''
    except subprocess.CalledProcessError as e:
        print(str(e))
        print(Stage + " failed")
        sys.exit(1)

def CheckDependencies():
    DependenciesMissing = False
    if shutil.which("mmseqs") is None:
        print("mmseqs not available")
        DependenciesMissing = True

    if shutil.which("psiblast") is None:
        print("psiblast not available")
        DependenciesMissing = True

    if shutil.which("blastdbcmd") is None:
        print("blastdbcmd not available")
        DependenciesMissing = True

    if shutil.which("python") is None:
        print("python not available")
        DependenciesMissing = True

    if shutil.which("muscle") is None:
        print("muscle not available")
        DependenciesMissing = True

    if DependenciesMissing:
        print("required software missing")
        sys.exit(1)

    return

ap = argparse.ArgumentParser(description = "ICITY pipeline. Provides analysis of protein neighborhoods around " +
                                           "specified seeds (coordinates for list of loci of interest), using " +
                                           "spatial information of ORFs and their protein sequrncrs, to search for " +
                                           "protein families associated to the seeds. Input informstion - config ot input with default?")

opts = ap.parse_args()
CheckDependencies()
Dir = config.DefenseSystem_Name + "_OUTPUT"
if not os.path.exists(Dir):
    os.mkdir(Dir)

# Run steps here
subprocess_call("Step 5: Extracting seeds", "python HashSeedExtract.py -c " + config.ICITY_CONFIG_INPUT["CdsPath"] +
                " -s " + config.ICITY_CONFIG_INPUT["SeedPath"] +
                " -o " + config.ICITY_CONFIG_TEMPORARYFILES["SeedsExtractedFileName"])


subprocess_call("Step 6: Selecting neighborhoods", "python FindNeighborhood.py -c " + config.ICITY_CONFIG_INPUT["CdsPath"] +
                " -s " + config.ICITY_CONFIG_TEMPORARYFILES["SeedsExtractedFileName"] +
                " -o " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityFileName"] +
                " -d " + str(config.ICITY_CONFIG_INPUT["NeighborhoodVicinitySize"]))

subprocess_call("Step 7: Collecting protein IDs", "grep -v \"===\" " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityFileName"] + " | cut -f1 | sort -u > " +
                config.ICITY_CONFIG_TEMPORARYFILES["VicinityIDsFileName"])

subprocess_call("Step 8: Fetching protein sequences", "blastdbcmd -db " + config.ICITY_CONFIG_INPUT["PathToDatabase"] +
                " -entry_batch " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityIDsFileName"] +
                " -long_seqids > " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityFASTAFileName"])

subprocess_call("Step 9: Clustering protein seqiences", "bash RunClust.sh " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityFASTAFileName"] + " " +
                str(config.ICITY_CONFIG_INPUT["PermissiveClusteringThreshold"]) + " " +
                config.ICITY_CONFIG_OUTPUT["VicinityClustersFileName"])

subprocess_call("Step 10: Making profiles", "python MakeProfiles.py -f " + config.ICITY_CONFIG_OUTPUT["VicinityClustersFileName"] +
                " -c " + config.ICITY_CONFIG_TEMPORARYFILES["ProfilesFolder"] +
                " -d " + config.ICITY_CONFIG_INPUT["PathToDatabase"])


subprocess_call("Step 11: Running PSI-BLAST for profiles", "python RunPSIBLAST.py -c " + config.ICITY_CONFIG_TEMPORARYFILES["ProfilesFolder"] +
               " -d " + config.ICITY_CONFIG_INPUT["PathToDatabase"] + " -t " + config.ICITY_CONFIG_INPUT["ThreadNum"]
                 )

subprocess_call("Step 12: Sorting blast hits", "python SortBLASTHitsInMemory.py -c " + config.ICITY_CONFIG_TEMPORARYFILES["ProfilesFolder"] +
                " -o " + config.ICITY_CONFIG_TEMPORARYFILES["SortedBLASTHitsFolder"] +
                " -p " + config.ICITY_CONFIG_INPUT["PTYFile"] +
                " -i " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityIDsFileName"] +
                " -s " + config.ICITY_CONFIG_TEMPORARYFILES["SeedsExtractedFileName"] +
                " -v " + config.ICITY_CONFIG_TEMPORARYFILES["VicinityFileName"] +
                " -z " + str(config.ICITY_CONFIG_INPUT["SortingOverlapThreshold"]) +
                " -x " + str(config.ICITY_CONFIG_INPUT["SortingCoverageThresold"]))

subprocess_call("Step 13: Calculating ICITY metric", "bash Cal.sh " +
                config.ICITY_CONFIG_TEMPORARYFILES["SortedBLASTHitsFolder"] + " " +
                config.ICITY_CONFIG_INPUT["PathToDatabase"] + " " +
                config.ICITY_CONFIG_OUTPUT["VicinityClustersFileName"] + " " +
                config.ICITY_CONFIG_OUTPUT["ICITYFileName"] + " " +
                config.ICITY_CONFIG_INPUT["ThreadNum"]
                )

subprocess_call("Step 14: Sorting Relevance", "python SortRelevance.py -n " + config.DefenseSystem_Name+
                " -p " + config.DefenseSystem_FilePath
                )



import os
import argparse
import subprocess


# Relevance.tsv
CLUSTER_NAME = 0
NUMBER_IN_VICINTY = 1
NUMBER_IN_ENTIRE = 2
DISTANCE_TO_SEED = 3
DS3P = 4



# 阈值暂定，需要使用统计得到一个有说服力的数值
DS3P_threshold = 0.0

# CLUSTER.hits_sorted

PROTEIN_ACCESSION = 0

# Seed.tsv
SEED_ACCESSION = 2


# Reading arguments from command line
ap = argparse.ArgumentParser(description = "SortRelevance")
ap.add_argument("-n", help = "DefenseSystemName", required = True)
ap.add_argument("-p", help = "DefenseSystemFilePath", required = True)

opts = ap.parse_args()



# 修改文件位置
DefenseSystem_Name = opts.n
DefenseSystem_FilePath = opts.p

SeedAccession_Filename = DefenseSystem_FilePath+ DefenseSystem_Name + '_OUTPUT/Seeds_' + DefenseSystem_Name + '.tsv'
CLUSTERS_Filepath = DefenseSystem_FilePath + DefenseSystem_Name + '_OUTPUT/CLUSTERS_' + DefenseSystem_Name
Vicinity_Filename = (os.path.split(DefenseSystem_FilePath+ DefenseSystem_Name + '_OUTPUT/Vicinity_'+ DefenseSystem_Name + '.faa')[1])
Relevance_Filenpath = DefenseSystem_FilePath + DefenseSystem_Name + '_OUTPUT/Relevance_' + DefenseSystem_Name + '.tsv'
Relevance_CategoryName = 'Relevance_Sorted_'+ DefenseSystem_Name + '_Category.csv'
CLUSTERHitsSorted_path = DefenseSystem_FilePath + DefenseSystem_Name + '_OUTPUT/CLUSTERS_' + DefenseSystem_Name + '/Sorted/'


DefenseSystem_Function_1 = 'abortive infection'
DefenseSystem_Function_2 = 'restriction-modification'
DefenseSystem_Function_3 = 'toxin'
DefenseSystem_Function_4 = 'argonaute'
DefenseSystem_Function_5 = 'Dnd'
Dir = "ACCESSION_A"

if not os.path.exists(Dir):
    os.mkdir(Dir)


subprocess.call("cat " + CLUSTERS_Filepath + "/CLUSTER_*.ali  | grep ref | awk -F \"|\" '{print $2, \"|\", $NF}'" + " > ACCESSION_A/ACCESSION_"+ DefenseSystem_Name+'.txt',
shell= True)
subprocess.call('cat ACCESSION_A/ACCESSION_'+ DefenseSystem_Name + '.txt | grep \''+ DefenseSystem_Function_1 +'\' | awk \'{print $0\" |  1\" }\' > ACCESSION_A/ACCESSION_ONLY_'+ DefenseSystem_Name +'.txt',
shell= True)
subprocess.call('cat ACCESSION_A/ACCESSION_' + DefenseSystem_Name + '.txt | grep -E \"'+ DefenseSystem_Function_2 + '|CRISPR|'+ DefenseSystem_Function_3 +'|'+DefenseSystem_Function_4+'|'+DefenseSystem_Function_5+'\" | awk \'{print $0  \" |  2\" }\' > ACCESSION_A/ACCESSION_Other_DefenseGene_'+ DefenseSystem_Name +'.txt',
shell= True)
subprocess.call('cat ACCESSION_A/ACCESSION_'+ DefenseSystem_Name +'.txt | grep -v \"toxin\|CRISPR\|abortive infection\|restriction-modification\|hypothetical\|argonaute\Dnd" | awk \'{print $0  \" |  3\" }\' > ACCESSION_A/ACCESSION_HouseKeepingGene_'+ DefenseSystem_Name + '.txt',
shell= True)
subprocess.call('cat ACCESSION_A/ACCESSION_'+ DefenseSystem_Name + '.txt | grep \'hypothetical\' | awk \'{print $0  \" |  4\" }\' > ACCESSION_A/ACCESSION_hypothetical_'+ DefenseSystem_Name + '.txt',
shell= True)




# 配合使用ACCESSION.sh 得到分类文件,不同抗性系统需要修改文件路径
Accession_All_FileName = 'ACCESSION_A/ACCESSION_' + DefenseSystem_Name + '.txt'
Accession_Category_Target_FileName = 'ACCESSION_A/ACCESSION_ONLY_'+ DefenseSystem_Name + '.txt'
Accession_Category_Hypothetical_FileName = 'ACCESSION_A/ACCESSION_hypothetical_'+ DefenseSystem_Name + '.txt'
ACCESSION_HouseKeepingGene_FileName = 'ACCESSION_A/ACCESSION_HouseKeepingGene_'+ DefenseSystem_Name + '.txt'
ACCESSION_Other_DefenseGene_FileName = 'ACCESSION_A/ACCESSION_Other_DefenseGene_'+ DefenseSystem_Name + '.txt'

with open(Relevance_Filenpath,'r') as F:
    Relevance = dict()
    for Line in F:
        LineValue = Line[:-1].split('\t')
        Relevance[LineValue[CLUSTER_NAME]] = [LineValue[CLUSTER_NAME], LineValue[NUMBER_IN_VICINTY], LineValue[NUMBER_IN_ENTIRE], LineValue[DISTANCE_TO_SEED], LineValue[DS3P]]

Relevance_Sorted_Target = dict()
for r in Relevance:
    if int(Relevance[r][NUMBER_IN_ENTIRE]) >= 1:
        if float(Relevance[r][DS3P]) > DS3P_threshold:
            Relevance_Sorted_Target[r] = Relevance[r]
    else:
        continue


for n in Relevance_Sorted_Target:
    CLUSTERHitsSorted_Filename = CLUSTERHitsSorted_path + n + '.hits_sorted'
    with open(CLUSTERHitsSorted_Filename,'r') as F:
        ProteinAccession = []
        for Line in F:
            LineValue = Line[:-1].split('\t')
            ProteinAccession.append(LineValue[PROTEIN_ACCESSION])
    
    Relevance_Sorted_Target[n] = [Relevance_Sorted_Target[n], ProteinAccession]


# 构建字典，引入分类符
# 与输入数据一致的防御基因--1
# 其他防御基因--2
# 管家基因--3
# 未知基因--4
Accession_Category_Target = dict()
Accession_Category_Hypothetical = dict()
Accession_Category_OtherDefense = dict()
Accession_Category_HouseKeeping = dict()

with open(Accession_Category_Target_FileName,'r') as T:
    for Line in T:
        LineValue = Line[:-1].split(' |  ')
        Accession_Category_Target[LineValue[0]] = [LineValue[1], LineValue[2]]

with open(ACCESSION_Other_DefenseGene_FileName,'r') as T:
    for Line in T:
        LineValue = Line[:-1].split(' |  ')
        Accession_Category_HouseKeeping[LineValue[0]] = [LineValue[1], LineValue[2]]

with open(ACCESSION_HouseKeepingGene_FileName,'r') as T:
    for Line in T:
        LineValue = Line[:-1].split(' |  ')
        Accession_Category_OtherDefense[LineValue[0]] = [LineValue[1], LineValue[2]]

with open(Accession_Category_Hypothetical_FileName,'r') as T:
    for Line in T:
        LineValue = Line[:-1].split(' |  ')
        Accession_Category_Hypothetical[LineValue[0]] = [LineValue[1], LineValue[2]]

Relevance_Sorted_Target_Category = dict()
for id in Relevance_Sorted_Target:
    # 每个Cluster 里多个 Accession 的功能可能不同，若不同，选取顺序如下: 1>2>3>4
    Accession_Category= {}
    for ac in Relevance_Sorted_Target[id][1]:
        if ac in Accession_Category_Target:
            Accession_Category[ac] = 1
            continue
        elif ac in Accession_Category_OtherDefense:
            Accession_Category[ac] = 2
            continue
        elif ac in Accession_Category_HouseKeeping:
            Accession_Category[ac] = 3
            continue
        elif ac in Accession_Category_Hypothetical:
            Accession_Category[ac] = 4
            continue
        else:
            Accession_Category[ac] = 5
    Cluster_Category_Name = min(Accession_Category, key= Accession_Category.get)
    Cluster_Category_Value = min(Accession_Category.values())
    if Cluster_Category_Value == 1:
        Relevance_Sorted_Target_Category[id] = [Relevance_Sorted_Target[id][0], Accession_Category_Target[Cluster_Category_Name][1]]
    elif Cluster_Category_Value == 2:
        Relevance_Sorted_Target_Category[id] = [Relevance_Sorted_Target[id][0], Accession_Category_OtherDefense[Cluster_Category_Name][1]]
    elif Cluster_Category_Value == 3:
        Relevance_Sorted_Target_Category[id] = [Relevance_Sorted_Target[id][0], Accession_Category_HouseKeeping[Cluster_Category_Name][1]]
    elif Cluster_Category_Value == 4:
        Relevance_Sorted_Target_Category[id] = [Relevance_Sorted_Target[id][0], Accession_Category_Hypothetical[Cluster_Category_Name][1]]


# 计算物种保守性（属-种分开算）
def CalConservation(Accession_Taxonomy):
        count = 0
        if not 'None' in Accession_Taxonomy:
            for key in Accession_Taxonomy.keys():
                count = count + 1/len(Accession_Taxonomy[key])
            return count/len(Accession_Taxonomy)
        else:
            for key in Accession_Taxonomy.keys():
                if key != 'None':
                    count = count + 1/len(Accession_Taxonomy[key])
            return count/len(Accession_Taxonomy)
    
# 引入物种分类信息（属-种）
Relevance_Taxonomy = dict()
Relevance_Sorted_Target_Taxonomy = dict()

with open(Accession_All_FileName,'r') as T:
    for Line in T:
        LineValue = Line[:-1].split(' |  ')
        TaxonomyID = LineValue[1].split('[')[1][:-1]
        TaxonomyID = TaxonomyID.split(' ',1)
        Relevance_Taxonomy[LineValue[0]] = TaxonomyID

for id in Relevance_Sorted_Target:
    Accession_Taxonomy_genus = {}
    Accession_Taxonomy_species = {}

    for ac in Relevance_Sorted_Target[id][1]:
        if ac in Relevance_Taxonomy:
            Accession_Taxonomy_genus[ac] = Relevance_Taxonomy[ac][0]
            # 部分Accession缺少species的注释
            try:
                Accession_Taxonomy_species[ac] = Relevance_Taxonomy[ac][1]
            except IndexError:
                Accession_Taxonomy_species[ac] = 'None'

        flipped_genus = {}
        flipped_species = {}

        for key,value in Accession_Taxonomy_genus.items():
            if value not in flipped_genus:
                flipped_genus[value] = list()
                flipped_genus[value].append(key)
            else:
                flipped_genus[value].append(key)

        for key,value in Accession_Taxonomy_species.items():   
            if value not in flipped_species:
                flipped_species[value] = list()
                flipped_species[value].append(key)
            else:
                flipped_species[value].append(key)
    if flipped_genus == {}:
        continue
    else:
        Relevance_Sorted_Target_Taxonomy[id] = [CalConservation(flipped_genus),CalConservation(flipped_species)]


with open(Relevance_CategoryName,'w') as f:
 for i in Relevance_Sorted_Target_Category:
    for element in Relevance_Sorted_Target_Category[i][0]:
        f.write(element)
        f.write(',')
    f.write(Relevance_Sorted_Target_Category[i][1])
    for n in Relevance_Sorted_Target_Taxonomy[i]:
        f.write(',')
        f.write(str(n))
        
    f.write("\n")





import setuptools
setuptools.setup(
    name="ICityRunner",
    version="0.1.1",
    author="Maxwell Wong",
    author_email="78146185@qq.com",
    description="ICity Runner for bioinfomatics",
    # long_description=open("README.md").read(),
    license="MIT",
   # url="https://github.com/Maxwell-Wong/ICityRunner",
    packages=['ICityRunner'],
    install_requires=[
        "numpy==1.19.2",
        "pandas==1.4.3"

        ],
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Unix Shell",
    ],





)


import time
import numpy as np
import pandas as pd
import argparse


# Hyper-parameters
i = 0
cnt = 0
read_time_e = 0
read_time_s = 0
t1 = 0
t2 = 0

# Reading arguments from command line
ap = argparse.ArgumentParser(description = "Extracting seeds from CDS file")
ap.add_argument("-s", help = "SeedsFileName, seeds xlsx file", required = True)
ap.add_argument("-c", help = "CDSFileName, cds pty file, containing seed information", required = True)
ap.add_argument("-o", help = "ResultFileName, output tsv file", required = True)
opts = ap.parse_args()

seed_path = opts.s  # Archaea_RM.xlsx
cds_path = opts.c  # archaea_cds.pty
seeds_name = opts.o   # Seeds_RM_A.tsv


# O(n)构建哈希表
def create_hash_table(cds_find):
    hash_table = dict()
    for idx, row in enumerate(cds_find.itertuples()):
        if getattr(row, 'cds_accession') not in hash_table.keys():
            hash_table.update({getattr(row, 'cds_accession'): idx + 1})
    return hash_table


def SeedExtract(seed_path, cds_path):
    global i, read_time_s, read_time_e, t1, t2
    info_alloc = pd.DataFrame()
    read_time_s = time.time()
    cds_file = pd.read_table(cds_path, names=['LociID', 'ORFStart_ORFStop',
                                              'Strand', 'OrganismID', 'ContigID', 'product_accession'])

    cds_file[['Start', 'Stop']] = cds_file["ORFStart_ORFStop"].str.split(':', n=1, expand=True)
    cds_file['Stop'] = cds_file['Stop'].str.replace('>', '')
    cds_file[['OrganismID', 'assembly_accession']] = cds_file['OrganismID'].str.rsplit('-', n=1, expand=True)
    cds_file = cds_file.drop(labels='ORFStart_ORFStop', axis=1)
    cds_file = cds_file.drop(labels='OrganismID', axis=1)

    cds_accession = cds_file.assembly_accession
    cds_end = cds_file.Stop
    cds_find = pd.concat([cds_accession, cds_end], axis=1); cds_find.columns = ['cds_accession', 'cds_end']

    hash_table = create_hash_table(cds_find)
    hash_table_values = list(hash_table.values())
    print("----------------------------HASH TABLE----------------------------",flush=True)
    print(hash_table, flush=True)
    print("HASH_TABLE LENGTH = ", len(hash_table), flush=True)

    cds_find = cds_find.to_numpy()

    seed_file = pd.read_excel(seed_path)
    seed_assembly_accession = seed_file.assembly_accession.str.replace("GCA", "GCF")
    seed_end = seed_file.end
    seed_find = pd.concat([seed_assembly_accession, seed_end], axis=1)

    read_time_e = time.time()

    print("------------------------Processing Starting Now------------------------", flush=True)
    t1 = time.time()
    for id, end in zip(seed_find.assembly_accession, seed_find.end):
        if id not in hash_table.keys():
            continue
        if hash_table_values.index(hash_table[id]) + 1 < len(hash_table_values):
            l2 = hash_table_values[hash_table_values.index(hash_table[id]) + 1]
        else:
            l2 = len(cds_file)
        l1 = int(hash_table[id])  # Starting Point
        length = l2 - l1
        print(id, ' ---------> Search length =', length, '[Start =', l1, ', Stop =', l2, ']', flush=True)
        accession_list = cds_find[int(hash_table[id]): int(hash_table[id]) + length + 1]

        for idx, item in enumerate(accession_list):
            if (item == np.array([id, str(end)])).all():
                print('No.', i, ' Seed Found: ', id, str(end), flush=True)
                # Process the corresponding items
                data = cds_file.loc[idx + l1, ['assembly_accession',
                                        'LociID', 'product_accession', 'ContigID', 'Start', 'Stop']]
                data = data.to_dict()
                data = pd.DataFrame(data, index=[i]); i += 1
                info_alloc = pd.concat([info_alloc, data])
        t2 = time.time()
    info_alloc.drop_duplicates(inplace=True)
    print("------------------------Processing Finished------------------------",flush=True)
    print(info_alloc,flush=True)
    print('----------------------------', i, 'Seeds in total----------------------------',flush=True)
    info_alloc.to_csv(seeds_name, index=False, sep='\t', header=False, )


# Hit it!
SeedExtract(seed_path, cds_path)
print("---------------------------------Performance Attributes---------------------------------")
print('File saved as ' + seeds_name)
print('Reading files and creating hash table used time = %.3fs' %(read_time_e - read_time_s))
print('Searching used time = %.3fs' %(t2 - t1))
print("Total used time = %.3fs" %(read_time_e - read_time_s + t2 - t1))
print("Seed Extraction Done!!")

